#include "../../../internal-include/CD/cd_internal.hpp"
#include <PSX/File/Processor/cd_file_processor.hpp>
#include <stdio.h>

namespace JabyEngine {
    static constexpr auto NormalCircularBufferSize   = 5;
    static constexpr auto DisabledCircularBufferSize = 512;

    void CDFileProcessor :: start_cur_job() {
        using CD::internal::SectorBufferAllocator;
            const auto configurate_for = [this](const CDFile& file) {
                const auto disable_lz4 = [this](uint32_t* work_area, size_t size, uint32_t* overwrite_dst = nullptr) -> uint32_t* {
                    uint8_t*const dst_adr = reinterpret_cast<uint8_t*>(this->circular_buffer.setup(reinterpret_cast<CD_IO::DataSector*>(work_area), size));

                    this->lz4_decomp.disable();
                    return overwrite_dst ? overwrite_dst : reinterpret_cast<uint32_t*>(dst_adr);
                };

                const auto enable_lz4 = [this](uint32_t* work_area, size_t size, uint32_t* override_dst = nullptr) -> uint32_t* {
                    uint8_t* dst_adr = reinterpret_cast<uint8_t*>(this->circular_buffer.setup(reinterpret_cast<CD_IO::DataSector*>(work_area), size));
                    
                    if(override_dst) {
                        dst_adr = reinterpret_cast<uint8_t*>(override_dst);
                    }

                    this->lz4_decomp.setup(dst_adr);
                    return reinterpret_cast<uint32_t*>(dst_adr);
                };

                switch(file.type) {
                    case CDFileType::SimpleTIM:
                        this->file_state = FileProcessor::create(enable_lz4(this->tmp_area, NormalCircularBufferSize), file.payload.simple_tim);
                        break;

                    case CDFileType::CopyTo:
                        this->file_state = FileProcessor::create(disable_lz4(file.payload.copy_to.dst, 512), Nothing());
                        break;

                    case CDFileType::Overlay:
                        this->file_state = FileProcessor::create(enable_lz4(this->tmp_area, NormalCircularBufferSize, file.payload.overlay.dst), Nothing());
                        break;
                }            
            };

        const auto& cur_job = *this->jobs.files;
        const auto& cur_lba = this->lba[cur_job.rel_lba_idx];
        
        configurate_for(cur_job);
        CD::internal::read_file(cur_lba, SectorBufferAllocator::create(this,
            [](void* ctx) -> CD_IO::DataSector* {
                CDFileProcessor &self = *reinterpret_cast<CDFileProcessor*>(ctx);
                return self.circular_buffer.allocate();
            }));

        printf(">>> CD needs to load LBA: %i -> %i\n", cur_lba.get_lba(), cur_lba.get_size_in_sectors());
    }

    bool CDFileProcessor :: process_data() {
        while(this->circular_buffer.has_data()) {
            ArrayRange<const uint8_t> cur_sector(reinterpret_cast<uint8_t*>(this->circular_buffer.get_next()->data), CD_IO::DataSector::SizeBytes);
            
            //                                                       v We can not know if there will be more data or not - but it also doesn't matter much for us
            const auto result = this->lz4_decomp.process(cur_sector, false);
            this->circular_buffer.pop();

            if(result) {
                // Process the data in the tmp_area
                if(this->file_state.process(result.bytes_ready) == Progress::Error) {
                    return false;
                }
            }

            else {
                return false;
            }
        }
        return true;
    }

    void CDFileProcessor :: setup(const volatile AutoLBAEntry* lba, JobArray jobs, uint32_t* tmp_area) {
        this->lba      = const_cast<const AutoLBAEntry*>(lba);
        this->jobs     = jobs;

        this->tmp_area = tmp_area;
        CDFileProcessor::start_cur_job();
    }

    Progress CDFileProcessor :: process() {
        const auto cur_state = CD::internal::read_current_state();

        CDFileProcessor::process_data();
        switch(cur_state) {
            case CD::internal::State::Done:
                /* 
                    We are done now!
                    The user decides if he wants the next value
                */
               //while(this->file_state.process(0) != Progress::Done);
                return Progress::Done;

            case CD::internal::State::BufferFull:
                /* We processd data and unpause the CD drive */
                CD::internal::continue_reading();
                return Progress::InProgress;

            case CD::internal::State::Reading:
                return Progress::InProgress;

            case CD::internal::State::Error:
            default:
                /* Error for real */
                return Progress::Error;
        }
    }
}